from __future__ import annotations
from typing import Dict, Any, Optional, Tuple, Iterator
from types import ModuleType
from pydantic import BaseModel, Field
from ghostos_common.entity import EntityMeta, to_entity_meta, from_entity_meta, is_entity_type

__all__ = [
    'PyContext',
]


class PyContext(BaseModel):
    """
    MOSS 运行依赖的 python 上下文.
    """

    module: Optional[str] = Field(
        default=None,
        description="the module path from which import the moss context predefined code",
    )
    code: Optional[str] = Field(
        default=None,
        description="if code given, use it instead of code from the module",
    )

    # injections: Dict[str, Injection] = Field(
    #     default_factory=dict,
    #     description="通过 python 引入的包, 类, 方法 等. 会注入到 MOSS 上, 同时会实现它.",
    # )

    properties: Dict[str, EntityMeta] = Field(
        default_factory=dict,
    )

    # properties: Dict[str, Property] = Field(
    #     default_factory=dict,
    #     description="在上下文中定义的变量. 会注入到 MOSS 上. 修改后也会保存到 pycontext 里. ",
    # )

    execute_code: Optional[str] = Field(
        default=None,
        description="the generated python code on this context",
    )

    executed: bool = Field(
        default=False,
        description="if the generated code is executed",
    )

    def set_prop(self, name: str, value: Any):
        self.properties[name] = to_entity_meta(value)

    def get_prop(self, name: str, module: Optional[ModuleType] = None) -> Any:
        if name not in self.properties:
            return None
        value = self.properties[name]
        return from_entity_meta(value, module)

    @staticmethod
    def allow_prop(value: Any) -> bool:
        if isinstance(value, BaseModel):
            return True
        elif isinstance(value, bool):
            return True
        elif isinstance(value, str):
            return True
        elif isinstance(value, int):
            return True
        elif isinstance(value, float):
            return True
        elif isinstance(value, list):
            return True
        elif isinstance(value, dict):
            return True
        elif is_entity_type(value):
            return True
        return False

    def iter_props(self, module: Optional[ModuleType] = None) -> Iterator[Tuple[str, Any]]:
        for name in self.properties:
            value = self.properties[name]
            yield name, from_entity_meta(value, module)

    def join(self, ctx: "PyContext") -> "PyContext":
        """
        合并两个 python context, 以右侧的为准. 并返回一个新的 PyContext 对象. 避免左向污染.
        """
        copied = self.model_copy(deep=True)
        if copied.module is None:
            copied.module = ctx.module
        if copied.code is None:
            copied.code = ctx.code
        if copied.execute_code is None:
            copied.execute_code = ctx.execute_code
            copied.executed = ctx.executed

        for key, val in ctx.properties.items():
            copied.properties[key] = val
        return copied

# class Injection(BaseModel):
#     """
#     from module import specific attribute then inject to MOSS Context
#     """
#     import_from: str = Field(
#         description="the imported module name or use module path pattern such as 'modulename:attr_name'",
#     )
#     alias: Optional[str] = Field(default=None, description="context attr alias for the imported value")
#
#     @classmethod
#     def reflect(cls, value: Any, alias: Optional[str] = None) -> "Injection":
#         """
#         reflect a value and generate Imported value.
#         :param value:
#         :param alias:
#         :return:
#         """
#         modulename = inspect.getmodule(value).__name__
#         if inspect.ismodule(value):
#             spec = None
#         else:
#             spec = getattr(value, '__name__', None)
#         import_from = join_import_module_and_spec(modulename, spec)
#         return Injection(
#             import_from=import_from,
#             alias=alias,
#         )
#
#     def get_name(self) -> str:
#         if self.alias:
#             return self.alias
#         _, spec = self.get_from_module_attr()
#         return spec
#
#     def get_from_module_attr(self) -> Tuple[str, Optional[str]]:
#         """
#         :return: modulename and attribute name from the module
#         """
#         return parse_import_module_and_spec(self.import_from)
#
#
# SerializableType = Union[str, int, float, bool, None, list, dict, BaseModel, TypedDict]
# """系统支持的各种可序列化类型, 可以被存储到 Serializable 里. 更多情况下可以用别的变量类型. """
# SerializableData = Union[str, int, float, bool, None, List, Dict]
#
#
# class Property(BaseModel):
#     """
#     可以在 MOSS 上下文中声明的变量.
#     """
#     name: str = Field(default="", description="property name in the moss context")
#     desc: str = Field(default="", description="describe the property's purpose")
#     value: Union[str, int, float, bool, None, list, dict, BaseModel, TypedDict] = Field(
#         default=None,
#         description="the serializable value",
#     )
#     model: Optional[str] = Field(
#         default=None,
#         description="如果是 pydantic 等类型, 可以通过类进行封装. 类应该在 imports 或者 defines 里.",
#     )
#
#     def __set_name__(self, owner, name):
#         self.name = name
#         if hasattr(owner, '__pycontext__') and isinstance(owner.__pycontext__, PyContext):
#             owner.__pycontext__.define(self)
#
#     def __set__(self, instance, value):
#         if value is self:
#             return
#         if isinstance(value, Property):
#             self.value = value.value
#             self.model = value.model
#             self.name = value.name
#             self.model = value.model
#         self.set_value(value)
#
#     def __get__(self, instance, owner):
#         return self.generate_value()
#
#     def __delete__(self):
#         self.__value__ = None
#         self.value = None
#         self.model = None
#
#     @classmethod
#     def from_value(cls, *, name: str = "", value: SerializableType, desc: str = "") -> Optional["Property"]:
#         p = cls(name=name, desc=desc)
#         p.set_value(value)
#         return p
#
#     def set_value(self, value: Any) -> None:
#         if not isinstance(value, SerializableType):
#             # 赋值的时候报错.
#             raise AttributeError(f"{value} is not property serializable type {SerializableType}")
#         model = None
#         has_model = (
#                 value is not None and isinstance(value, BaseModel)
#                 or is_typeddict(value)
#         )
#         if has_model:
#             type_ = type(value)
#             if type_.__qualname__:
#                 model = type_.__module__ + ':' + type_.__qualname__
#             else:
#                 model = type_.__module__ + ':' + type_.__name__
#         self.value = value
#         self.model = model
#
#     def generate_value(self, module: Optional[ModuleType] = None) -> Any:
#         model = self.model
#         value = self.value
#         if isinstance(self.value, dict) and model is not None:
#             if not isinstance(value, Dict):
#                 raise AttributeError(f"'{value}' is not dict while model class is '{model}'")
#             cls = None
#             if module is not None:
#                 modulename, spec = parse_import_module_and_spec(model)
#                 if modulename == module.__name__:
#                     # 用这种方法解决临时模块里的变量问题.
#                     cls = get_module_spec(module.__dict__, spec)
#             if cls is None:
#                 cls = import_from_path(model)
#             if issubclass(cls, BaseModel):
#                 self.value = cls(**value)
#         return self.value
#
#
# @definition()
# def attr(
#         default: SerializableType = None, *,
#         default_factory: Optional[Callable[[], Any]] = None,
#         desc: str = "",
# ) -> SerializableType:
#     """
#     用于定义一个要绑定到 MOSS 上的属性, 它的值可以在多轮对话和思考过程中保存和修改.
#     :param default: 属性的默认值, 目前支持 str, int, float, bool, None, list, dict, BaseModel, TypedDict
#     :param default_factory: 可以传入一个 lambda 或闭包, 当 default 为 None 时生成值.
#     :param desc: 属性的描述.
#     """
#     if default is None and default_factory:
#         default = default_factory()
#     return Property.from_value(value=default, desc=desc)



from typing import List, Dict, Optional, Callable
from ghostos_moss.abcd import MossCompiler, Execution
from ghostos_moss.pycontext import PyContext
from ghostos_container import Container
from queue import Queue
from threading import Thread


class MossTestSuite:
    MAGIC_TEST_CASES_ATTR_NAME = "__moss_test_cases__"

    def __init__(self, container: Container):
        self._container = container

    def container(self) -> Container:
        return self._container

    def dump_prompt(
            self,
            *,
            modulename: str,
            test_modulename: str = "__test__",
    ) -> str:
        compiler = self._container.force_fetch(MossCompiler)
        compiler.join_context(PyContext(module=modulename))
        runtime = compiler.compile(test_modulename)
        return runtime.prompter().dump_module_prompt()

    def run_module_tests(
            self, *,
            modulename: str,
            callback: Callable[[str, Execution], None],
            test_modulename: str = "__test__",
            targets: Optional[str] = None,
    ) -> None:
        """
        run the test cases in moss module, and returns each case's result to callback function.
        :param callback: callback on (test_case_name, MossResult)
        :param modulename: the target moss module
        :param test_modulename: the modulename that MossCompiler shall build.
        :param targets: the target functions that should be tested,
                        if None, test the func names from `__moss_test_cases__`
        :return:
        """
        compiler = self._container.force_fetch(MossCompiler)
        compiler.join_context(PyContext(module=modulename))
        runtime = compiler.compile(test_modulename)
        compiled = runtime.module()
        if not targets:
            targets: List[str] = compiled.__dict__.get(self.MAGIC_TEST_CASES_ATTR_NAME)
            if not isinstance(targets, List):
                raise AttributeError(f"Module {modulename} has no {self.MAGIC_TEST_CASES_ATTR_NAME} attribute")
        if not targets:
            raise AttributeError(f"test cases are empty")
        self.parallel_run_moss_func(
            modulename=modulename,
            callback=callback,
            test_module_name=test_modulename,
            funcs=targets,
        )

    def run(
            self, *,
            modulename: str,
            test_module_name: str = "__test__",
            target: str = "test_main",
            args: Optional[List[str]] = None,
            kwargs: Dict[str, str] = None,
    ) -> Execution:
        """
        运行一个指定的 moss 测试.
        :param modulename: 想要测试的 moss 文件的模块路径.
        :param test_module_name: 运行时创建的临时 module 名.
        :param target: 想要调用 moss 文件中的某个方法.
        :param args: 给 target 方法传递的 args 参数. 会从 module.__dict__ 查找真值.
        :param kwargs: 给 target 方法传递的 kwargs. 会从 module.__dict__ 查找真值.
        :return: Moss 执行后的 result.
        """
        compiler = self._container.force_fetch(MossCompiler)
        compiler.join_context(PyContext(module=modulename))
        runtime = compiler.compile(test_module_name)
        result = runtime.execute(target=target, local_args=args, local_kwargs=kwargs)
        return result

    def parallel_run_moss_func(
            self, *,
            modulename: str,
            funcs: List[str],
            callback: Callable[[str, Execution], None],
            test_module_name: str = "__test__",
    ) -> None:
        """
        并发地运行多个测试方法, FIFO 返回结果.
        :param callback: callback on (str, MossResult)
        :param modulename: 目标 module 的名字.
        :param funcs: 需要运行的 funcs.
        :param test_module_name: 测试时创建的临时 module_name.
        :return: 从队列中逐个取出.
        """
        queue = Queue()

        def runner(fn: str, q: Queue) -> None:
            """
            测试用 queue 来阻塞返回结果.
            :param fn: 要测试的方法名.
            :param q: 队列
            :param c: condition
            :param
            """
            r = self.run(
                modulename=modulename,
                test_module_name=test_module_name,
                target=fn,
                args=['moss'],
            )
            q.put((fn, r))

        threads = []
        for func in funcs:
            t = Thread(target=runner, args=(func, queue))
            t.start()
            threads.append(t)

        i = 0
        while i < len(threads):
            result = queue.get(block=True)  # 获取数据
            name, moss_result = result
            callback(name, moss_result)
            i += 1

        queue.task_done()
        for t in threads:
            t.join()



from abc import ABC
from typing import Iterable, Tuple, Any
from ghostos_moss.prompts import reflect_code_prompt, join_prompt_lines
from ghostos_moss.utils import add_source_indent, escape_string_quotes
import inspect

__all__ = ['Exporter', 'is_exports']


class Exporter(ABC):
    """
    class that exports function, method and class of a module.
    and moss.get_imported_attr_prompts will reflect all the attributes assign to the Exports subclass.

    How to:

    1. module `Foo` define an Exports subclass for other module to import:
    ```python
    from ghostos_moss import Exports
    class Foo:
        ...

    class Exports(AbsExports):
        Foo = Foo
    ```

    2. module `Bar` import the Exports and use it:

    ```python
    from foo import Exports

    my_foo = Exports.Foo()
    ```

    3. MossPrompter.get_imported_attr_prompts() can get the exported attributes code interfaces from it:

    ```python
    with compiler:
        runtime = compiler.compile("bar")
        assert `class Exports` in runtime.prompter().get_imported_attr_prompts()
    ```
    """

    @classmethod
    def __attrs_prompts__(cls) -> Iterable[Tuple[str, str]]:
        """
        override this method to define str prompt for attribute
        :return: Iterable[(attr_name, attr_prompt)]
        """
        return []

    @classmethod
    def __exports__(cls) -> Iterable[Tuple[str, Any]]:
        """
        iterate exports attributes
        """
        attrs = dir(cls)
        for attr in attrs:
            if attr.startswith("_"):
                continue
            value = getattr(cls, attr, None)
            if value is None:
                continue
            yield attr, value

    @classmethod
    def __class_prompt__(cls) -> str:
        if cls is Exporter:
            return ""
        source = inspect.getsource(cls)
        attr_prompts = {}
        for prop, prompt in cls.__attrs_prompts__():
            attr_prompts[prop] = prompt

        for attr, value in cls.__exports__():
            try:
                attr_interface = reflect_code_prompt(value)
            except Exception:
                continue

            if attr_interface:
                attr_prompts[attr] = attr_interface

        blocks = []
        for name, prompt in attr_prompts.items():
            value = getattr(cls, name, None)
            if value is None:
                continue
            module_info = ""
            if hasattr(value, "__module__"):
                module_info = f" module=`{value.__module__}`"

            lines = [
                f"#<attr name=`{name}`{module_info}>",
                prompt,
                f"#</attr>",
            ]
            blocks.append("\n".join(lines))
        if not blocks:
            return source
        self_name = cls.__name__
        appending = join_prompt_lines(*blocks)
        appending = escape_string_quotes(appending, '"""')
        appending = "\n".join([f'"""\n#attrs of `{self_name}` are:', appending, '"""'])
        appending = add_source_indent(appending, 4)
        return join_prompt_lines(source.rstrip(), appending)


def is_exports(obj) -> bool:
    return issubclass(obj, Exporter)


class _ExportsExample(Exporter):
    """
    example of Exports.
    the source code and
    """
    getsource = inspect.getsource
    Exports = Exporter



from abc import ABC, abstractmethod
from typing import Any, Optional, Dict, Tuple, Iterable, Union, Callable
from typing_extensions import Protocol, is_typeddict
from types import ModuleType
from ghostos_moss.utils import (
    get_modulename,
    get_callable_definition,
)
from pydantic import BaseModel
from dataclasses import is_dataclass
import inspect

"""
将上下文引用的 变量/方法/类型 反射出 Prompt 的机制. 
主要解决一个问题, 如何让一个属性能够被大模型所理解. 

本质上有三种机制: 
+ 类: 展示折叠的, 或者全部的源码. 
+ 方法: 展示折叠的, 或者全部的源码. 
+ 属性: 展示属性的 typehint. 又有几种做法: 
    - 赋值: 类似 `x:int=123` 的形式展示. 
    - 类型: 没有赋值, 只有 `x: foo` 的方式展示. 
    - 字符串类型: 用字符串的方式来描述类型. 比如 `x: "<foo.Bar>"`. 其类型说明是打印结果. 
    - doc: 在 python 的规范里, 属性可以在其下追加字符串作为它的说明. 

预计有以下几种机制: 

1. 在代码里手写注释或者字符串说明. 
2. 如果变量拥有 __prompt__ 属性, 通过它 (可以是方法或字符串) 生成 prompt. 
"""

__all__ = [
    'reflect_code_prompt',
    'reflect_locals_imported', 'reflect_class_with_methods',
    'join_prompt_lines', 'compile_attr_prompts',
    'AttrPrompts',
    'PromptAbleClass', 'PromptAbleObj',
]

AttrPrompts = Iterable[Tuple[str, str]]
"""
描述多个属性的代码, 作为 prompt 提供给 LLM. 
每个元素代表一个属性. 
元素的值为 Tuple[name, prompt] 
name 是为了去重, prompt 应该就是原有的 prompt. 

如果 attr_name 存在, 使用 f"{name}{prompt}" 格式化, 预计结构是:`name[: typehint] = value[\n\"""doc\"""]`
如果 attr_name 不存在, 则直接使用 prompt. 

多条 prompt 用 "\n\n".join(prompts) 的方式拼接. 
"""

ignore_modules = {
    "pydantic",
}


def get_defined_prompt(value: Any) -> Union[str, None]:
    attr = get_defined_prompt_attr(value)
    if attr is None:
        return None
    if isinstance(attr, str):
        return attr
    return attr()


def get_defined_prompt_attr(value: Any) -> Union[None, str, Callable[[], str]]:
    if value is None:
        return None
    elif isinstance(value, PromptAbleObj):
        return value.__prompt__

    elif isinstance(value, type):
        if issubclass(value, PromptAbleClass):
            return value.__class_prompt__
        # class without __class_prompt__ is not defined as prompter
        if hasattr(value, "__class_prompt__"):
            return getattr(value, "__class_prompt__")

    elif hasattr(value, "__prompt__"):
        prompter = getattr(value, "__prompt__")
        if inspect.isfunction(value) or inspect.ismethod(value) or hasattr(prompter, '__self__'):
            return prompter
    elif isinstance(value, ModuleType) and '__prompt__' in value.__dict__:
        prompter = value.__dict__['__prompt__']
        return prompter
    return None


class PromptAbleObj(ABC):
    """
    拥有 __prompt__ 方法的类.
    这里只是一个示范, 并不需要真正继承这个类, 只需要有 __prompt__ 方法或属性.
    """

    @abstractmethod
    def __prompt__(self) -> str:
        pass


class PromptAbleProtocol(Protocol):
    @abstractmethod
    def __prompt__(self) -> str:
        pass


class PromptAbleClass(ABC):

    @classmethod
    @abstractmethod
    def __class_prompt__(cls) -> str:
        pass


class PromptAbleClassProtocol(Protocol):

    @classmethod
    @abstractmethod
    def __class_prompt__(cls) -> str:
        pass


PromptAble = Union[PromptAbleClass, PromptAbleObj, PromptAbleProtocol, PromptAbleClassProtocol]


# todo: change function name
def reflect_locals_imported(
        modulename: str,
        local_values: Dict[str, Any],
) -> AttrPrompts:
    """
    MOSS 系统自带的反射方法, 对一个module 的本地变量做最小化的反射展示.
    基本原理:
    1. 当前模块变量:
       - 当前模块的变量默认不展示, 因为本地变量的 prompt 可以直接写在代码里.
       - 如果定义了 __prompt__ 方法, 则会展示出来.
    2. 不反射任何 `_` 开头的本地变量.
    3. 不反射 builtin 类型.
    4. 如果目标是 module
        - 包含 __prompt__ 方法时嵌套展示
        - 否则不展示. 避免递归问题.
    5. 如果目标是 function
        - 包含 __prompt__ 方法时使用它生成,
        - 否则返回 function 的 definition + doc
    6. 如果目标是 class
        - 包含 __class_prompt__ 方法时, 用它生成.
        - __is_abstract__ 的 class, 直接返回源码.
    7. 如果目标是其它 attr
        _ 只有包含 prompt 方法时才展示.

    :param modulename: 当前模块名. 所有当前模块的变量默认不展示.
    :param local_values: 传入的上下文变量.
    """
    for name, value in local_values.items():
        try:
            prompt = reflect_imported_attr(name, value, modulename)
        except Exception as e:
            raise RuntimeError(f"failed to reflect local value {name!r}: {e}")
        if prompt is not None:
            yield name, prompt


# todo: change function name
def reflect_class_with_methods(cls: type) -> str:
    """
    reflect class with all its method signatures.
    """
    from inspect import getsource
    from ghostos_moss.utils import make_class_prompt, get_callable_definition
    source = getsource(cls)
    attrs = []
    for name in dir(cls):
        if name.startswith("_"):
            continue
        method = getattr(cls, name)
        if inspect.ismethod(method) or inspect.isfunction(method):
            block = get_callable_definition(method)
            attrs.append(block)
    return make_class_prompt(source=source, attrs=attrs)


def reflect_imported_attr(
        name: str,
        value: Any,
        current_module: str,
) -> Optional[str]:
    """
    反射其中的一个值.
    """
    if name.startswith('_'):
        # 私有变量不展示.
        return None
    if inspect.isbuiltin(value):
        # 系统内置的, 都不展示.
        return None

    # module 相关的过滤逻辑.
    value_modulename = get_modulename(value)
    if value_modulename is None:
        return None
    elif value_modulename == current_module:
        return None
    for ignore_module_name in ignore_modules:
        if value_modulename.startswith(ignore_module_name):
            return None

    return reflect_code_prompt(value)


def reflect_code_prompt(value: Any) -> Optional[str]:
    """
    get prompt from value.
    only:
    1. predefined PromptAble
    2. abstract class
    3. function or method
    will generate prompt
    """
    prompt = get_defined_prompt(value)
    if prompt is not None:
        return prompt

    if inspect.isclass(value):
        # only reflect abstract class
        if inspect.isabstract(value) or issubclass(value, BaseModel) or is_dataclass(value) or is_typeddict(value):
            source = inspect.getsource(value)
            if source:
                return source
    elif inspect.isfunction(value) or inspect.ismethod(value):
        # 默认都给方法展示 definition.
        return get_callable_definition(value)

    return None


def join_prompt_lines(*prompts: Optional[str]) -> str:
    """
    将多个可能为空的 prompt 合并成一个 python 代码风格的 prompt.
    """
    result = []
    for prompt in prompts:
        line = prompt.rstrip()
        if line:
            result.append(prompt)
    return '\n\n'.join(result)


def compile_attr_prompts(attr_prompts: AttrPrompts) -> str:
    prompts = []
    for name, prompt in attr_prompts:
        prompt = prompt.strip()
        if not prompt:
            continue
        attr_prompt = f'''# <attr name=`{name}`>
{prompt}
# </attr>'''
        prompts.append(attr_prompt)
    return join_prompt_lines(*prompts)



import importlib
import inspect
from types import ModuleType
from typing import Optional, Any, Dict, get_type_hints, Type, List, Callable, ClassVar
import io

from ghostos_container import Container, Provider
from ghostos.contracts.modules import Modules, ImportWrapper
from ghostos_moss.abcd import (
    Moss,
    MossCompiler, MossRuntime, MossPrompter, MOSS_VALUE_NAME, MOSS_TYPE_NAME,
    MOSS_HIDDEN_MARK, MOSS_HIDDEN_UNMARK,
    Injection,
)
from ghostos_moss.prompts import reflect_code_prompt
from ghostos_moss.pycontext import PyContext
from ghostos_moss.exports import Exporter
from ghostos_moss.magics import replace_magic_prompter
from ghostos_common.helpers import generate_module_and_attr_name, code_syntax_check, get_code_interface_str
from contextlib import contextmanager, redirect_stdout

IMPORT_FUTURE = "from __future__ import annotations"

__all__ = [
    'MossStub',
    'MossCompilerImpl',
    'MossRuntimeImpl',
    'DefaultMOSSProvider',
    'MossTempModuleType',
]


class MossTempModuleType(ModuleType):
    __instance_count__: ClassVar[int] = 0

    def __del__(self):
        if MOSS_VALUE_NAME in self.__dict__:
            del self.__dict__[MOSS_VALUE_NAME]
        MossTempModuleType.__instance_count__ -= 1


class MossCompilerImpl(MossCompiler):
    def __init__(self, *, container: Container, pycontext: Optional[PyContext] = None):
        self._container = Container(parent=container, name="moss")
        self._pycontext = pycontext if pycontext else PyContext()
        self._modules: Modules = container.force_fetch(Modules)
        self._predefined_locals: Dict[str, Any] = {
            # 默认代理掉 import.
            '__import__': ImportWrapper(self._modules),
        }
        self._injections: Dict[str, Any] = {}
        self._attr_prompts: List = []
        # the default ignored modules
        self._ignored_modules = [
            'pydantic',
            'typing',
            'typing_extensions',
            'inspect',
            'abc',
            'openai',
        ]
        self._compiled = False
        self._closed = False

    def __del__(self):
        self.close()

    def container(self) -> Container:
        return self._container

    def pycontext(self) -> PyContext:
        return self._pycontext

    def join_context(self, context: PyContext) -> "MossCompiler":
        self._pycontext = self._pycontext.join(context)
        return self

    def with_locals(self, **kwargs) -> "MossCompiler":
        self._predefined_locals.update(kwargs)
        return self

    def with_ignored_imported(self, module_names: List[str]) -> "MossCompiler":
        self._ignored_modules = module_names
        return self

    def injects(self, **attrs: Any) -> "MossCompiler":
        self._injections.update(attrs)
        return self

    def _compile(self, modulename: Optional[str] = None) -> ModuleType:
        origin: Optional[ModuleType] = None
        filename = "<moss_temp_module>"
        origin_modulename = self._pycontext.module
        # 没有任何设定的情况下.
        if origin_modulename is None and self._pycontext.code is None:
            # 认为传入的 modulename 就是要编译的 module
            self._pycontext.module = modulename
            origin_modulename = modulename

        self._container.bootstrap()
        if origin_modulename:
            origin = importlib.import_module(origin_modulename)
            filename = origin.__file__

        # 编译的 module name.
        if modulename is None:
            modulename = origin_modulename if origin_modulename else "__moss__"

        code = self.pycontext_code()
        # 创建临时模块.
        module = MossTempModuleType(modulename)
        MossTempModuleType.__instance_count__ += 1
        module.__dict__.update(self._predefined_locals)
        module.__file__ = filename
        compiled = compile(code, modulename, "exec")
        exec(compiled, module.__dict__)
        if origin is not None:
            updating = self._filter_origin(origin)
            module.__dict__.update(updating)
        return module

    @staticmethod
    def _filter_origin(origin: ModuleType) -> Dict[str, Any]:
        result = {}
        for attr_name, attr_value in origin.__dict__.items():
            if attr_name.startswith("__"):
                continue
            if inspect.isclass(attr_value) or inspect.isfunction(attr_value):
                if attr_value.__module__ != origin.__name__:
                    continue
                result[attr_name] = attr_value
        return result

    def _new_runtime(self, module: ModuleType) -> "MossRuntime":
        attr_prompts = {}
        for attr_name, value in self._attr_prompts:
            attr_prompts[attr_name] = value
        self._compiled = True
        return MossRuntimeImpl(
            container=self._container,
            pycontext=self._pycontext.model_copy(deep=True),
            source_code=self.pycontext_code(),
            compiled=module,
            injections=self._injections,
            attr_prompts=attr_prompts,
            ignored_modules=self._ignored_modules,
        )

    def pycontext_code(self) -> str:
        code = self._pycontext.code
        module = self._pycontext.module
        if code is None:
            if module is None:
                return ""
            module = self._modules.import_module(self._pycontext.module)
            code = inspect.getsource(module)
        if not code.lstrip().startswith(IMPORT_FUTURE):
            code = IMPORT_FUTURE + "\n\n" + code.lstrip("\n")
        return code if code else ""

    def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        # container 先不 destroy.
        if not self._compiled:
            self._container.shutdown()
        del self._container
        del self._pycontext
        del self._predefined_locals
        del self._injections


class MossStub(Moss):
    instance_count: ClassVar[int] = 0
    __pycontext__: PyContext
    __container__: Container
    __output__: str
    __printer__: Callable

    def __init__(
            self,
            pycontext: PyContext,
            container: Container,
            printer: Callable,
            watching: List,
            ignored: List,
    ):
        MossStub.instance_count += 1
        self.__dict__['__pycontext__'] = pycontext
        self.__dict__['__container__'] = container
        self.__dict__['__output__'] = ""
        self.__dict__['__printer__'] = printer
        self.__dict__['__watching__'] = watching
        self.__dict__['__ignored__'] = ignored

    def fetch(self, abstract: Moss.T) -> Optional[Moss.T]:
        return self.__container__.fetch(abstract)

    def pprint(self, *args, **kwargs) -> None:
        return self.__printer__(*args, **kwargs)

    def __setattr__(self, _name, _value):
        if self.__pycontext__.allow_prop(_value):
            self.__pycontext__.set_prop(_name, _value)
        self.__dict__[_name] = _value

    def __del__(self):
        MossStub.instance_count -= 1


def new_moss_stub(cls: Type[Moss], container: Container, pycontext: PyContext, pprint: Callable) -> Moss:
    # cls 必须不包含参数.

    watching = cls.__watching__
    stub = MossStub(pycontext, container, pprint, watching, cls.__ignored__)
    stub.executing_code = None
    # assert stub.instance_count > 0
    for attr_name in dir(cls):
        if not attr_name.startswith("_") and not hasattr(stub, attr_name):
            attr_value = getattr(cls, attr_name)
            setattr(stub, attr_name, attr_value)

    # 反向注入.
    for name, value in cls.__dict__.items():
        if name in pycontext.properties or name.startswith("_"):
            continue
        if stub.__pycontext__.allow_prop(value):
            stub.__pycontext__.set_prop(name, value)

    return stub


class MossRuntimeImpl(MossRuntime, MossPrompter):

    def __init__(
            self, *,
            container: Container,
            pycontext: PyContext,
            source_code: str,
            compiled: ModuleType,
            injections: Dict[str, Any],
            attr_prompts: Dict[str, str],
            ignored_modules: List[str],
    ):
        self._container = container
        self._modules: Modules = container.force_fetch(Modules)
        self._compiled = compiled
        self._source_code = source_code
        self._pycontext = pycontext
        self._injections = injections
        self._runtime_std_output = ""
        # 初始化之后不应该为 None 的值.
        self._built: bool = False
        self._moss_prompt: Optional[str] = None
        self._attr_prompts: Dict[str, str] = attr_prompts
        self._imported_attrs: Optional[Dict[str, Any]] = None
        self._closed: bool = False
        self._injected = set()
        self._injected_types = set()
        self._moss: Moss = self._compile_moss()
        self._initialize_moss()
        self._ignored_modules = set(ignored_modules) | set(self._moss.__ignored__)

        self._replaced_magic_prompts = replace_magic_prompter(self._compiled)
        MossRuntime.instance_count += 1

    def _compile_moss(self) -> Moss:
        moss_type = self.moss_type()
        if not issubclass(moss_type, Moss):
            raise TypeError(f"Moss type {moss_type} is not subclass of {generate_module_and_attr_name(Moss)}")

        # 创建 stub.
        pycontext = self._pycontext
        moss = new_moss_stub(moss_type, self._container, pycontext, self.pprint)
        return moss

    def _initialize_moss(self) -> None:
        from .lifecycle import __moss_compiled__
        moss = self._moss
        pycontext = self._pycontext
        moss_type = self.moss_type()

        def inject(attr_name: str, injected: Any) -> Any:
            if isinstance(injected, Injection):
                injected.on_inject(self, attr_name)
            setattr(moss, attr_name, injected)
            self._injected.add(attr_name)

        # 初始化 pycontext variable
        for name, prop in pycontext.iter_props(self._compiled):
            # 直接用 property 作为值.
            inject(name, prop)

        # 反向注入

        for name, injection in self._injections.items():
            inject(name, injection)

        # 初始化基于容器的依赖注入.
        typehints = get_type_hints(moss_type, localns=self._compiled.__dict__)
        for name, typehint in typehints.items():
            if name.startswith('_'):
                continue

            # 记录所有定义过的类型.
            self._injected_types.add(typehint)

            # 已经有的就不再注入.
            if hasattr(moss, name):
                continue

            # 为 None 才依赖注入.
            value = self._container.force_fetch(typehint)
            # 依赖注入.
            inject(name, value)

        self._compiled.__dict__[MOSS_VALUE_NAME] = moss
        self._compiled.__dict__[MOSS_TYPE_NAME] = moss_type
        fn = __moss_compiled__
        if __moss_compiled__.__name__ in self._compiled.__dict__:
            fn = self._compiled.__dict__[__moss_compiled__.__name__]
        fn(moss)
        self._moss = moss

    def container(self) -> Container:
        return self._container

    def prompter(self) -> MossPrompter:
        return self

    def lint_exec_code(self, code: str) -> Optional[str]:
        source_code = self._source_code
        new_code = source_code + "\n\n" + code.strip()
        return code_syntax_check(new_code)

    def module(self) -> ModuleType:
        return self._compiled

    def locals(self) -> Dict[str, Any]:
        return self._compiled.__dict__

    def moss(self) -> Moss:
        return self._moss

    def dump_pycontext(self) -> PyContext:
        return self._pycontext

    def dump_std_output(self) -> str:
        return self._runtime_std_output

    def pprint(self, *args: Any, **kwargs: Any) -> None:
        from pprint import pprint
        out = io.StringIO()
        with redirect_stdout(out):
            pprint(*args, **kwargs)
        self._runtime_std_output += str(out.getvalue())

    @contextmanager
    def redirect_stdout(self):
        buffer = io.StringIO()
        with redirect_stdout(buffer):
            yield
            self._runtime_std_output += str(buffer.getvalue())

    def get_source_code(
            self,
            exclude_hide_code: bool = True,
    ) -> str:
        code = self._source_code
        return self._parse_pycontext_code(code, exclude_hide_code)

    def moss_injections(self) -> Dict[str, Any]:
        moss = self.moss()
        injections = {}
        for name in self._injected:
            injection = getattr(moss, name)
            injections[name] = injection
        return injections

    def replaced_magic_prompts(self) -> Dict[str, str]:
        return self._replaced_magic_prompts

    def get_imported_attrs(self) -> Dict[str, Any]:
        if self._imported_attrs is None:
            self._imported_attrs = {}
            self_modulename = self._compiled.__module__
            for name, value in self._compiled.__dict__.items():
                if name.startswith("_"):
                    continue
                elif not value:
                    continue
                elif inspect.isbuiltin(value):
                    # 系统内置的, 都不展示.
                    continue
                elif inspect.ismodule(value):
                    # module can not get itself
                    self._imported_attrs[name] = value
                elif hasattr(value, "__module__") and hasattr(value, "__name__"):
                    value_module = value.__module__
                    if value_module == self_modulename:
                        continue
                    self._imported_attrs[name] = value
        return self._imported_attrs

    def _is_ignored(self, modulename: str) -> bool:
        if modulename in self._ignored_modules:
            return True
        for ignored in self._ignored_modules:
            ignored_prefix = ignored.rstrip('.') + '.'
            if modulename.startswith(ignored_prefix):
                return True
        return False

    def get_imported_attrs_prompt(self, includes: List = None) -> str:
        # todo: make build class later
        imported_attrs = self.get_imported_attrs()
        reverse_imported = {}
        reflection_types = {Moss}

        # add all function and method to reflections.
        # and prepare reverse imported.
        for key, value in imported_attrs.items():
            if value not in reverse_imported:
                reverse_imported[value] = key
                # only function or methods are reflected automatically
                if inspect.isfunction(value) or inspect.ismethod(value):
                    reflection_types.add(value)
                elif inspect.isclass(value) and issubclass(value, Exporter):
                    # Exports class always add to reflect types.
                    reflection_types.add(value)
        if includes:
            for value in includes:
                if value not in reverse_imported:
                    continue
                reflection_types.add(value)

        # add injected first .
        for injected_type in self._injected_types:
            # only imported value shall be reflected.
            if injected_type in reverse_imported:
                reflection_types.add(injected_type)

        for watched in self.moss().__watching__:
            if watched in reverse_imported:
                reflection_types.add(watched)

        blocks = []
        functions = []
        classes = []
        modules = []
        module_names = set()
        others = []
        for reflection_type in reflection_types:
            if inspect.isclass(reflection_type):
                classes.append(reflection_type)
            elif inspect.ismodule(reflection_type):
                modules.append(reflection_type)
                module_names.add(reflection_type.__name__)
            elif inspect.isroutine(reflection_type):
                functions.append(reflection_type)
            else:
                others.append(reflection_type)
        if len(functions) > 0:
            blocks.append("#<routines>")
            for item in functions:
                name = reverse_imported[item]
                item_module = item.__module__
                if self._is_ignored(item_module) or item_module in module_names:
                    continue
                prompt = reflect_code_prompt(item)
                if not prompt:
                    continue
                name_desc = f" name=`{name}`" if name != item.__name__ else ""
                if name_desc:
                    block = f"#<local{name_desc}>\n{prompt}\n#</local>"
                else:
                    block = prompt
                blocks.append(block)
            blocks.append("#</routines>")
        if len(classes) > 0:
            blocks.append("#<classes>")
            for item in classes:
                name = reverse_imported[item]
                item_module = item.__module__
                if self._is_ignored(item_module) or item_module in module_names:
                    continue
                name_desc = f" name=`{name}`" if name != item.__name__ else ""
                prompt = reflect_code_prompt(item)
                if not prompt:
                    continue
                if name_desc:
                    block = f"#<local{name_desc}>\n{prompt}\n#</local>"
                else:
                    block = prompt
                blocks.append(block)
            blocks.append("#</classes>")
        if len(modules) > 0:
            blocks.append("#<modules>")
            for item in modules:
                name = reverse_imported[item]
                item_module = item.__name__
                if self._is_ignored(item_module):
                    continue
                source = inspect.getsource(item)
                if not source:
                    continue
                prompt = get_code_interface_str(source)
                if prompt:
                    block = f"#<local name=`{name}` module=`{item_module}`>\n{prompt}\n#</local>"
                    blocks.append(block)
            blocks.append("#</modules>")
        if len(others) > 0:
            blocks.append("#<others>")
            for item in others:
                name = reverse_imported[item]
                prompt = reflect_code_prompt(item)
                if not prompt:
                    prompt = inspect.getsource(item)
                if prompt:
                    type_ = type(item).__name__
                    block = f"#<local name=`{name}` type=`{type_}`>\n{prompt}\n# </local>"
                    blocks.append(block)
            blocks.append("#</others>")

        return "\n\n".join(blocks)

    @staticmethod
    def _parse_pycontext_code(code: str, exclude_hide_code: bool = True) -> str:
        if not exclude_hide_code:
            return code

        lines = code.split("\n")
        results = []
        num_hidden_mark = 0
        hide = False
        for line in lines:
            if line.lstrip().startswith(MOSS_HIDDEN_MARK):
                num_hidden_mark += 1
                hide = True
            elif line.lstrip().startswith(MOSS_HIDDEN_UNMARK):
                num_hidden_mark -= 1
                if num_hidden_mark <= 0:
                    hide = False
            else:
                if hide:
                    continue
                else:
                    results.append(line)

        return "\n".join(results)

    def update_executing_code(self, code: Optional[str] = None) -> None:
        if code is None:
            return
        self._pycontext.execute_code = code
        self._moss.executing_code = code

    def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        data = self._moss.__dict__
        for val in data.values():
            if isinstance(val, Injection):
                val.on_destroy()
        self._container.shutdown()

    def __del__(self):
        if not self._closed:
            self.close()
        MossRuntime.instance_count -= 1
        del self._container
        del self._injections
        del self._compiled
        del self._moss
        del self._pycontext


class DefaultMOSSProvider(Provider[MossCompiler]):
    """
    用于测试的标准 compiler.
    但实际上好像也是这个样子.
    """

    def singleton(self) -> bool:
        return False

    def contract(self) -> Type[MossCompiler]:
        return MossCompiler

    def factory(self, con: Container) -> MossCompiler:
        return MossCompilerImpl(container=con, pycontext=None)



from ghostos_moss.abcd import Moss


def plus(a: int, b: int) -> int:
    return a + b


# <moss-hide>
if __name__ == '__test__':
    """
    可以这样定义只在当前文件编译成 modulename=__test__ 才运行的方法. 
    """


    def test_1(moss: Moss) -> int:
        return plus(0, 1)


    def test_2(moss: Moss) -> int:
        return plus(1, 1)


    def test_3(moss: Moss) -> int:
        return plus(1, 2)


    __moss_test_cases__ = ['test_1', 'test_2', 'test_3']
    """用这个魔术变量, 可以让 MossTestSuit 批量调用三个方法测试. """

# </moss-hide>



from abc import ABC, abstractmethod
from typing import List

from ghostos_container import Container
from ghostos_moss.abcd import Moss as Parent
from ghostos.prompter import BasePOM
from inspect import getmembers, getsource
from pydantic import BaseModel


class Foo(ABC):
    """定义一个本地类, 用来做依赖注入测试. """

    @abstractmethod
    def foo(self) -> str:
        pass


def plus(a: int, b: int) -> int:
    """ 验证本地方法存在于 prompt. """
    return a + b


class TestPrompter(BasePOM):
    line: str = "TestPrompter"

    def self_prompt(self, container: Container) -> str:
        return self.line

    def get_title(self) -> str:
        return ""


class Moss(Parent, ABC):
    """
    本地定义的 Moss 类. 每个 MOSS 文件里都应该有一个 Moss 类, 可以是 import 的也可以是本地定义的.
    记得它要继承自 Moss.
    """
    life: List[str] = []
    """测试 attr 方法用来定义可持久化的属性. """

    foo: Foo
    """依赖注入 Foo 的测试用例. """

    tester: TestPrompter


# <moss-hide>
# !!! 使用 `# <moss-hide>` 和 `# </moss-hide>` 包裹的代码不会对大模型呈现.

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ghostos_moss.abcd import MossCompiler, AttrPrompts, MossPrompter, Execution


def __moss_compile__(compiler: "MossCompiler") -> "MossCompiler":
    """
    从 compile 中获取 MOSSRuntime, 并对它进行初始化.
    可选的魔术方法. 如果定义的话, MOSS 执行 compile 的阶段会默认执行它.

    主要解决各种注入方面的需求:
    """
    # 单测里应该有这个. moss.bar == 123
    compiler.injects(bar=123, tester=TestPrompter())
    # 插入生命周期事件, 直接赋值到 moss 上.
    Moss.life.append("__moss_compile__")

    # 准备依赖注入 Foo.
    class FooImpl(Foo):
        def foo(self) -> str:
            return "hello"

    from ghostos_container import provide
    # 用这种方式, 可以预期 Moss 被依赖注入了 Foo, 注入的是 FooImpl
    provider = provide(Foo, singleton=False)(lambda con: FooImpl())
    compiler.register(provider)

    return compiler


def __moss_attr_prompts__() -> "AttrPrompts":
    return [
        # 重写了 getsource 的 prompt, 它就应该不存在了.
        ("getsource", ""),
        # 添加一个意义不明的注释. 也应该在 prompt 里.
        ("9527", "# hello world")
    ]


def __moss_prompt__(prompter: "MossPrompter") -> str:
    # 测试生命周期生效.
    from ghostos_moss.lifecycle import __moss_module_prompt__
    return __moss_module_prompt__(prompter)


def __moss_exec__(*args, **kwargs) -> "Execution":
    # 测试生命周期生效.
    from ghostos_moss.lifecycle import __moss_exec__
    return __moss_exec__(*args, **kwargs)


def test_main(moss: Moss) -> int:
    """
    模拟一个 main 方法, 测试 moss 的调用.
    assert 返回值是 3. 外部的 MOSSRuntime 调用这个方法.
    """
    print("hello world")
    return plus(1, 2)


if __name__ == "__test__":
    def main(moss: Moss) -> int:
        """
        模拟一个 main 方法, 测试 moss 的调用.
        assert 返回值是 4. 外部的 MOSSRuntime 调用这个方法.
        """
        return plus(2, 2)

# </moss-hide>



import inspect
import re
from typing import Any, Callable, Optional, List, Iterable
from typing_extensions import is_typeddict
from pydantic import BaseModel

__all__ = [

    'unwrap_str',
    'get_modulename',

    'is_typing', 'is_builtin', 'is_classmethod',
    'is_model_class',
    'parse_comments',
    'parse_doc_string', 'escape_string_quotes',
    'strip_source_indent', 'add_source_indent', 'make_class_prompt',
    'is_callable', 'is_public_callable', 'get_callable_definition',
    'get_typehint_string', 'get_import_comment', 'get_extends_comment',
    'get_class_def_from_source',
    'count_source_indent',
    'replace_class_def_name',
    'get_calling_modulename',
    'is_code_same_as_print',
    'is_name_public',
    'add_comment_mark',
]


def get_import_comment(module: Optional[str], module_spec: Optional[str], alias: Optional[str]) -> Optional[str]:
    if module:
        if module_spec:
            if alias and alias != module_spec:
                return f"# from {module} import {module_spec} as {alias}"
            else:
                return f"# from {module} import {module_spec}"
        elif alias and not module.endswith(alias):
            return f"# import {module} as {alias}"
        else:
            return f"# import {module}"
    return None


def get_extends_comment(extends: Optional[List[Any]]) -> Optional[str]:
    if not extends:
        return None
    result = []
    for imp in extends:
        if not imp:
            continue
        elif isinstance(imp, str):
            result.append(imp)
        else:
            result.append('"' + str(imp) + '"')
    return "# extends " + ", ".join(result)


def get_typehint_string(typehint: Optional[Any]) -> str:
    if not typehint:
        return ""
    if isinstance(typehint, str):
        if typehint.lstrip().startswith(":"):
            return typehint
        return ": " + typehint
    if is_typing(typehint):
        return ": " + str(typehint)
    else:
        return ': "' + str(typehint) + '"'


def parse_doc_string(doc: Optional[str], inline: bool = True, quote: str = '"""') -> str:
    if not doc:
        return ""
    gap = "" if inline else "\n"
    doc = strip_source_indent(doc)
    doc = escape_string_quotes(doc, quote=quote)
    return quote + gap + doc + gap + quote


def parse_comments(comment: Optional[str]) -> str:
    if not comment:
        return ""
    comments = comment.split('\n')
    result = []
    for c in comments:
        c = c.strip()
        if not c.startswith('#'):
            c = '# ' + c
        result.append(c)
    return '\n'.join(result)


def make_class_prompt(
        *,
        source: str,
        name: Optional[str] = None,
        doc: Optional[str] = None,
        attrs: Optional[Iterable[str]] = None,
) -> str:
    source = strip_source_indent(source)
    class_def = get_class_def_from_source(source)
    if name:
        class_def = replace_class_def_name(class_def, name)
    if doc:
        doc = parse_doc_string(doc, inline=False, quote='"""')
    if doc:
        class_def += "\n" + add_source_indent(doc, 4)
    blocks = []
    if attrs:
        for attr in attrs:
            blocks.append(attr)
    if len(blocks) == 0:
        class_def += "\n" + add_source_indent("pass", 4)
        return class_def

    i = 0
    for block in blocks:
        _block = add_source_indent(block, 4)
        exp = "\n\n" if i > 0 else "\n"
        class_def += exp + _block
        i += 1
    return class_def


def replace_class_def_name(class_def: str, new_name: str) -> str:
    found = re.search(r'class\s+\w+[(:]', class_def)
    if not found:
        raise ValueError(f"Could not find class definition in {class_def}")
    found_str = found.group(0)
    found_str = found_str[:len(found_str) - 1]
    replace = f"class {new_name}"
    return class_def.replace(found_str, replace, 1)


def get_class_def_from_source(source: str) -> str:
    result = []
    source = strip_source_indent(source)
    source = source.strip()
    lines = source.split('\n')
    found_class = False
    for line in lines:
        line = line.rstrip()
        result.append(line)
        if line.startswith('class '):
            found_class = True
        unmarked = line.split('#')[0].rstrip()
        if found_class and unmarked.endswith(':'):
            break
    return '\n'.join(result)


def is_typing(value: Any) -> bool:
    if hasattr(value, "__module__"):
        return getattr(value, "__module__") == "typing"
    return False


def is_builtin(value: Any) -> bool:
    if inspect.isbuiltin(value):
        return True
    if not inspect.isclass(value):
        return False
    return value.__module__ == "__builtin__"


def is_classmethod(func: Any) -> bool:
    """
    判断一个变量是不是一个 @classmethod.
    code by moonshot
    """
    if not isinstance(func, Callable):
        return False
    if not inspect.ismethod(func):
        return False
    if not hasattr(func, '__self__'):
        return False
    self = getattr(func, '__self__', None)
    return self is not None and isinstance(self, type)


def unwrap_str(value: Any) -> Optional[str]:
    if isinstance(value, Callable):
        return value()
    try:
        return str(value)
    except AttributeError:
        return None


def get_callable_definition(
        caller: Callable,
        alias: Optional[str] = None,
        doc: Optional[str] = None,
) -> str:
    """
    将一个 callable 对象的源码剥离方法和描述.
    # todo: 用 tree-sitter 重做.
    """
    if doc:
        doc = doc.strip()
    if not inspect.isfunction(caller) and not inspect.ismethod(caller):
        if not inspect.isclass(caller) and isinstance(caller, Callable) and hasattr(caller, '__call__'):
            if not alias:
                alias = type(caller).__name__
            if not doc:
                doc = inspect.getdoc(caller)
            caller = getattr(caller, '__call__')
        else:
            raise TypeError(f'"{caller}" is not function or method')

    try:
        source_code = inspect.getsource(caller)
    except OSError:
        # 无法取到代码.
        return ""

    stripped_source = strip_source_indent(source_code)
    source_lines = stripped_source.split('\n')
    definition = []

    # 获取 method def
    for line in source_lines:
        # if line.startswith('def ') or len(definition) > 0:
        line = line.rstrip()
        if line == "@abstractmethod":
            continue
        definition.append(line)
        code_line = line.split('#')[0]
        if code_line.rstrip().endswith(':'):
            break
    defined = '\n'.join(definition).strip()
    if alias:
        found = re.search(r'def\s+(\w+)\(', defined)
        if found:
            defined = defined.replace(found.group(0), "def {name}(".format(name=alias), 1)
    indent_str = ' ' * 4
    if doc is None:
        doc = caller.__doc__ or ""
    if doc:
        doc = parse_doc_string(doc, inline=False)
        doc = add_source_indent(doc, indent=4)
        defined = defined + "\n" + doc
    defined = defined + "\n" + indent_str + "pass"
    return defined.strip()


def add_source_indent(source: str, indent: int = 4) -> str:
    """
    给代码添加前缀
    """
    source = source.rstrip()
    lines = source.split('\n')
    result = []
    indent_str = ' ' * indent
    for line in lines:
        if line.strip():
            line = indent_str + line
        result.append(line)
    return "\n".join(result)


def strip_source_indent(source_code: str, indent: Optional[int] = None) -> str:
    """
    一个简单的方法, 用来删除代码前面的 indent.
    """
    if indent is None:
        indent = count_source_indent(source_code)
    if indent == 0:
        return source_code
    indent_str = ' ' * indent
    source_lines = source_code.split('\n')
    result_lines = []
    for line in source_lines:
        if line.startswith(indent_str):
            line = line[indent:]
        result_lines.append(line)
    return '\n'.join(result_lines)


def count_source_indent(source_code: str) -> int:
    """
    一个简单的方法, 用来判断一段 python 函数代码的 indent.
    """
    source_lines = source_code.split('\n')
    for line in source_lines:
        right_stripped = line.rstrip()
        if len(right_stripped) == 0:
            continue
        both_stripped = right_stripped.lstrip()
        return len(right_stripped) - len(both_stripped)
    return 0


def escape_string_quotes(target: str, quote='"""') -> str:
    if target.startswith(quote) and target.endswith(quote):
        return target
    target = target.strip(quote)
    target = target.replace('\\' + quote, quote)
    target = target.replace(quote, '\\' + quote)
    return target.strip()


def add_name_to_set(names: set, name: str) -> set:
    if name in names:
        raise NameError(f'name "{name}" is already defined')
    names.add(name)
    return names


def is_model_class(typ: type) -> bool:
    if not isinstance(typ, type) or inspect.isabstract(typ):
        return False
    return issubclass(typ, BaseModel) or is_typeddict(typ)


def is_callable(obj: Any) -> bool:
    return isinstance(obj, Callable)


def is_name_public(name: str) -> bool:
    return not name.startswith('_')


def is_public_callable(attr: Any) -> bool:
    return isinstance(attr, Callable) and not inspect.isclass(attr) and not attr.__name__.startswith('_')


def get_calling_modulename(skip: int = 0) -> Optional[str]:
    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return None
    frame = stack[start][0]

    # module and packagename.
    module_info = inspect.getmodule(frame)
    if module_info:
        mod = module_info.__name__
        return mod
    return None


def is_code_same_as_print(value: Any) -> bool:
    return isinstance(value, bool) \
        or isinstance(value, int) \
        or isinstance(value, float) \
        or isinstance(value, complex)
    # or isinstance(value, list)
    # or isinstance(value, dict)


def get_modulename(val: Any) -> Optional[str]:
    name = getattr(val, '__module__', None)
    if name:
        return name
    module = inspect.getmodule(val)
    if module and hasattr(module, '__name__'):
        return getattr(module, '__name__', None)
    return None


def add_comment_mark(text: str, comment: str = "# ") -> str:
    lines = text.split('\n')
    contents = []
    for line in lines:
        if line.startswith(comment):
            contents.append(line)
        else:
            contents.append(comment + line)
    return "\n".join(contents)



from abc import ABC, abstractmethod
from typing import Type, Dict
from types import ModuleType

__all__ = [
    'is_magic_prompter', 'MagicPrompter', 'replace_magic_prompter',
    '__is_instance__', '__is_subclass__',
]


def __is_instance__(type_: Type):
    """
    list all the attrs of current module which are instances of type_
    """
    return _IsTypePrompter(type_)


def __is_subclass__(type_: Type):
    """
    list all the attrs of current module which are subclasses of type_
    """
    return _IsSubclassPrompter(type_)


class MagicPrompter(ABC):

    @abstractmethod
    def __moss_magic_prompt__(self, compiled: ModuleType) -> str:
        pass


def is_magic_prompter(value) -> bool:
    return isinstance(value, MagicPrompter)


def replace_magic_prompter(module: ModuleType) -> Dict[str, str]:
    replaced = {}
    for attr, value in module.__dict__.items():
        if isinstance(value, MagicPrompter):
            replaced[attr] = value.__moss_magic_prompt__(module)
    module.__dict__.update(replaced)
    return replaced


class _IsTypePrompter(MagicPrompter):

    def __init__(self, _type: type):
        self._type = _type

    def __moss_magic_prompt__(self, compiled: ModuleType) -> str:
        matched = []
        for name, value in compiled.__dict__.items():
            if name.startswith("_"):
                continue
            if isinstance(value, self._type):
                matched.append(name)

        return f'''
the attrs which is instance of {self._type} are:{", ".join(matched)}
'''


class _IsSubclassPrompter(MagicPrompter):

    def __init__(self, parent: type):
        self._parent = parent

    def __moss_magic_prompt__(self, compiled: ModuleType) -> str:
        matched = []
        for name, value in compiled.__dict__.items():
            if name.startswith("_"):
                continue
            if isinstance(value, type) and issubclass(value, self._parent):
                matched.append(name)
        return f'''
the attrs which is subclass of {self._parent} are: {", ".join(matched)}
'''



from __future__ import annotations
from typing import Dict, Any, Union, List, Optional, NamedTuple, Type, Callable, TypeVar, ClassVar
from typing_extensions import Self
from types import ModuleType, FunctionType
from abc import ABC, abstractmethod
from ghostos_container import Container, Provider, Factory, provide
from ghostos_moss.pycontext import PyContext
from ghostos_moss.prompts import (
    AttrPrompts, reflect_locals_imported, compile_attr_prompts
)

"""
MOSS 是 Model-oriented Operating System Simulation 的简写. 
它将系统当前上下文的 API 通过全代码的方式 Prompt 给模型, 让模型直接生成代码并且执行. 

MOSS 通过 PyContext 来定义一个可持久化, 可以复原的上下文. 
可以在多轮对话, 或者多轮思考中, 反复还原这个上下文. 

使用 moss 的完整生命周期的伪代码如下 (系统会自动集成)

# 第一步, 获取 MOSSCompiler 对象. 
compiler: MOSSCompiler = container.force_fetch(MOSSCompiler)

# 第二步, compiler 可以预定义 Module 的一些内部方法比如 __import__, 或 join pycontext, 然后 compile
# 关于 __import__, 默认会使用 contracts.modules.Modules 的实现. 

runtime: MOSSRuntime = compiler.join_pycontext(pycontext).compile(modulename)

# 这个阶段会编译 pycontext.module 到一个临时的 Module 中, 并对生成的 MOSSRuntime 进行初始化. 
# 如果 pycontext.code 存在, 则使用这个 code. 这是为了支持未来 llm 直接修改上下文, 存一个自己的临时副本. 

# 第三步, 生成 context 的 prompt. 
prompt = runtime.prompter().dump_context_prompt()

# 第四步, 使用 prompt 生成 chat, 调用大模型或人工生成代码. 
code = llm_runner(prompt)

# 第五步, 将 llm 生成的 code 让 moss runtime 执行, 并指定返回值或调用的方法 (target) 
result: MOSSResult = runtime.execute(code, target=xxxx)

# 第六步, 使用返回值, 并且 destroy()
returns = result.returns
std_output = result.std_output
# 获取需要保存的 pycontext. 
pycontext = result.pycontext

最后记得执行 runtime.destroy(), 否则可能内存泄漏. 
"""

__all__ = [
    'Moss',
    'MossCompiler', 'MossRuntime',
    'Execution', 'MossPrompter',
    # 'moss_message',
    'AttrPrompts',
    'MOSS_TYPE_NAME', 'MOSS_VALUE_NAME',
    'MOSS_HIDDEN_MARK', 'MOSS_HIDDEN_UNMARK',
    'Injection',
]

MOSS_TYPE_NAME = "Moss"

MOSS_VALUE_NAME = "moss"

MOSS_HIDDEN_MARK = "# <moss-hide>"
""" pycontext.module 源码某一行以这个标记开头, 其后的代码都不生成到 prompt 里. """

MOSS_HIDDEN_UNMARK = "# </moss-hide>"
""" pycontext.module 源码某一行以这个标记开头, 其后的代码都展示到 prompt 里. """


class Moss(ABC):
    """
    Language Model-oriented Operating System Simulator.
    Python interface of Runtime Injections for AI-Models in multi-turns chat or thinking.
    * The members with typehint will be injected with runtime instances.
    * The property of SerializeType will persist during multi-turns.
    * SerializeType: int, float, str, None, list, dict, BaseModel, TypedDict
    """

    T = TypeVar('T')

    executing_code: Optional[str]
    """the code that execute the moss instance."""

    __watching__: List[Union[FunctionType, ModuleType, type]] = []
    """the class or module that dose not bound to moss but still want to watch the interface of them"""

    __ignored__: List[str] = []
    """the ignored module names that do not need to watch the code interface of them"""


class Injection(ABC):

    @abstractmethod
    def on_inject(self, runtime: MossRuntime, property_name: str) -> Self:
        pass

    @abstractmethod
    def on_destroy(self) -> None:
        pass


class MossCompiler(ABC):
    """
    language Model-oriented Operating System Simulation
    full python code interface for large language models
    """

    # --- 创建 MOSS 的方法 --- #
    @abstractmethod
    def container(self) -> Container:
        """
        MOSS 专用的 IoC 容器.
        用来给 moss 实例实现依赖注入.
        """
        pass

    @abstractmethod
    def pycontext(self) -> PyContext:
        """
        MOSS 的动态上下文, 可以序列化保存到历史消息里, 从而在分布式系统里重新还原上下文.
        """
        pass

    @abstractmethod
    def join_context(self, context: PyContext) -> "MossCompiler":
        """
        为 MOSS 上下文添加一个可以持久化存储的 context 变量, 合并默认值.
        :param context:
        :return:
        """
        pass

    @abstractmethod
    def pycontext_code(self) -> str:
        """
        返回通过 pycontext.module 预定义的代码.
        """
        pass

    @abstractmethod
    def with_locals(self, **kwargs) -> "MossCompiler":
        """
        人工添加一些 locals 变量, 到目标 module 里. 在编译之前就会加载到目标 ModuleType.
        主要解决 __import__ 之类的内置方法, 如果必要的话.
        :param kwargs: locals
        :return: self
        """
        pass

    @abstractmethod
    def with_ignored_imported(self, module_names: List[str]) -> "MossCompiler":
        """
        ignored module names will not be prompt from imported attrs
        """
        pass

    def register(self, provider: Provider) -> None:
        """
        向生成 MOSS 的 IoC 容器里注册 Provider.
        当 MOSS 的属性定义了
        :param provider:
        :return:
        """
        self.container().register(provider)

    def bind(self, abstract: Any, value: Union[Factory, Any]) -> None:
        """
        向生成 MOSS 的 IoC 容器里注册工厂方法或实现.
        :param abstract: 抽象定义, 用来获取 value.
        :param value: 可以是一个值, 或者是一个 Factory. 如果是 Factory 的话, 取值的时候会使用它来做实例化.
        # todo: 具体实现可以记录日志.
        """
        if isinstance(value, Callable):
            provider = provide(abstract, singleton=False)(value)
            self.container().register(provider)
        else:
            self.container().set(abstract, value)

    @abstractmethod
    def injects(
            self,
            **attrs: Any,
    ) -> "MossCompiler":
        """
        inject certain values to the MOSS.
        Will automatically bind certain attributes to the MOSS instance (setattr).
        :param attrs: attr_name to attr_value (better with prompt)
        """
        pass

    __compiling__: bool = False
    __compiled__: bool = False

    def compile(
            self,
            modulename: Optional[str],
    ) -> "MossRuntime":
        """
        正式编译出一个 MOSSRuntime. 每一个 Compiler 只能编译一次.
        :param modulename: 生成的 ModuleType 所在的包名. 相关代码会在这个临时 ModuleType 里生成. 如果 modulename为空, 则使用原来的
        """
        from ghostos_moss.lifecycle import __moss_compile__
        if self.__compiling__:
            raise RuntimeError('recursively calling compile method')
        if self.__compiled__:
            raise RuntimeError('compiler already compiled')
        self.__compiling__ = True
        try:
            # 使用 locals 和 pycontext.module 对应的代码, 生成 ModuleType.
            module = self._compile(modulename)
            # 在生成的 ModuleType 里查找魔术方法, 提供 provider 等, 为依赖注入做准备.
            if hasattr(module, __moss_compile__.__name__):
                # 完成编译 moss_compiled 事件.
                # 可以在这个环节对 MOSS 做一些依赖注入的准备.
                fn = getattr(module, __moss_compile__.__name__)
                fn(self)
            runtime = self._new_runtime(module)
            return runtime
        finally:
            self.__compiling__ = False
            self.__compiled__ = True
            # 手动管理一下, 避免外部解决内存泄漏的心智成本.
            self.close()

    @abstractmethod
    def _compile(self, modulename: Optional[str] = None) -> ModuleType:
        """
        运行 pycontext.module 的代码, 编译 Module.
        """
        pass

    @abstractmethod
    def _new_runtime(self, module: ModuleType) -> "MossRuntime":
        """
        实例化 Runtime.
        :param module:
        :return:
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        主动做垃圾回收的准备, 避免 python 内存泄漏.
        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class MossPrompter(ABC):
    """
    MOSS 的运行时, 已经完成了第一阶段的编译.
    pycontext.module 的相关代码已经加载.
    这个阶段应该完成 MOSS 本身的注入准备.
    """

    @abstractmethod
    def container(self) -> Container:
        """
        MOSS 使用的依赖注入容器.
        """
        pass

    @abstractmethod
    def module(self) -> ModuleType:
        """
        当前的临时 Module.
        """
        pass

    @abstractmethod
    def get_source_code(
            self,
            exclude_hide_code: bool = True,
    ) -> str:
        """
        返回通过 pycontext.module 预定义的代码.
        第一行应该是 from __future__ import annotations. 解决上下文乱续的提示问题.
        :param exclude_hide_code: 如果为 True, 只返回大模型可以阅读的代码.
        """
        pass

    @abstractmethod
    def get_imported_attrs(self) -> Dict[str, Any]:
        """
        get imported attrs (function, class, or module) of the compiled module
        """
        pass

    @abstractmethod
    def get_imported_attrs_prompt(self, includes: List = None) -> str:
        """
        get the default prompt of the imported attrs.
        :return:
        """
        pass

    @abstractmethod
    def replaced_magic_prompts(self) -> Dict[str, str]:
        """
        the ghostos_moss.magics.MagicPrompter in this compiled module are replaced into str.
        :return: the replaced Dict[attr_name, prompt]
        """
        pass

    def get_magic_prompt(self) -> str:
        prompts = self.replaced_magic_prompts()
        if prompts:
            lines = [f"`{key}`:\n{prompt.strip()}" for key, prompt in prompts.items()]
            return "\n\n".join(lines)
        return ""

    def imported_attr_prompts(self) -> AttrPrompts:
        """
        结合已编译的本地变量, 用系统自带的方法反射出上下文属性的 prompts.
        """
        module = self.module()
        name = module.__name__
        local_values = module.__dict__
        yield from reflect_locals_imported(
            name,
            local_values,
        )

    def dump_imported_prompt(self, auto_generation: bool = True, attr_prompts: Dict[str, str] = None) -> str:
        """
        基于 pycontext code 生成的 Prompt. 用来描述当前上下文里的各种变量.
        主要是从其它库引入的变量.
        :return: 用 python 风格描述的上下文变量.
        """
        from ghostos_moss.lifecycle import __moss_attr_prompts__
        done = attr_prompts if attr_prompts else {}
        names = []
        # 查看是否有源码自带的魔术方法.
        module = self.module()
        if hasattr(module, __moss_attr_prompts__.__name__):
            fn = getattr(module, __moss_attr_prompts__.__name__)
            predefined_prompts: AttrPrompts = fn()
            for name, prompt in predefined_prompts:
                if prompt is None:
                    continue
                if name and name not in done:
                    names.append(name)
                done[name] = prompt

        # 合并系统自动生成的.
        if auto_generation:
            reflected_attr_prompts = self.imported_attr_prompts()
            for name, prompt in reflected_attr_prompts:
                if name not in done:
                    names.append(name)
                    done[name] = prompt

        # 保证一下顺序.
        prompts = [(name, done[name]) for name in names]
        return compile_attr_prompts(prompts)

    def dump_module_prompt(self, attr_prompts: Dict[str, str] = None) -> str:
        """
        获取 MOSS 运行时的完整 Python context 的 Prompt.
        这个 Prompt 包含以下几个部分:
        1. predefined_code: 通过 pycontext.module 默认加载的代码 (对大模型可展示的部分).
        2. pycontext_code_prompt: 对 predefined code 里各种引用类库的描述 prompt. 会包裹在 `\"""` 中展示.
        3. moss_prompt: moss 会注入到当前上下文里, 因此会生成 MOSS Prompt.
        """
        from ghostos_moss.lifecycle import __moss_module_prompt__
        compiled = self.module()
        # 基于 moss prompter 来生成.
        if hasattr(compiled, __moss_module_prompt__.__name__):
            fn = getattr(compiled, __moss_module_prompt__.__name__)
            return fn(self)
        return __moss_module_prompt__(self, attr_prompts)


class MossRuntime(ABC):
    instance_count: ClassVar[int] = 0

    @abstractmethod
    def container(self) -> Container:
        """
        MOSS 使用的依赖注入容器.
        """
        pass

    @abstractmethod
    def lint_exec_code(self, code: str) -> Optional[str]:
        """
        lint execution code and return error info if error occurs
        """
        pass

    @abstractmethod
    def prompter(self) -> MossPrompter:
        """
        返回一个 Prompter, 专门提供 prompt.
        """
        pass

    @abstractmethod
    def module(self) -> ModuleType:
        """
        最终运行的 Module.
        """
        pass

    @abstractmethod
    def locals(self) -> Dict[str, Any]:
        """
        返回运行时的上下文变量.
        其中一定包含 MOSS 类和注入的 moss 实例.
        """
        pass

    @abstractmethod
    def moss_injections(self) -> Dict[str, Any]:
        """
        get injections from moss
        """
        pass

    @abstractmethod
    def moss(self) -> Moss:
        """
        基于上下文生成的 MOSS. 依赖注入已经完成.
        """
        pass

    def moss_type(self) -> Type[Moss]:
        """
        get defined MOSS type
        :return: MOSS class
        """
        module = self.module()
        if MOSS_TYPE_NAME in module.__dict__:
            moss_type = module.__dict__[MOSS_TYPE_NAME]
            if not issubclass(moss_type, Moss):
                raise TypeError(f"Moss type {moss_type} is not subclass of {Moss}")
            return moss_type
        return Moss

    @abstractmethod
    def dump_pycontext(self) -> PyContext:
        """
        返回当前的可存储上下文.
        """
        pass

    @abstractmethod
    def dump_std_output(self) -> str:
        """
        返回重定向 buff 后的 std output
        需要结合 runtime_ctx 来开启 buffer.
        """
        pass

    @abstractmethod
    def redirect_stdout(self):
        """
        with runtime.exec_ctx():
            ...
        :return: a context object that co-operates with `with` statement
        """
        pass

    __executing__: bool = False

    def execute(
            self,
            *,
            target: str,
            code: Optional[str] = None,
            local_args: Optional[List[str]] = None,
            local_kwargs: Optional[Dict[str, str]] = None,
            args: Optional[List[Any]] = None,
            kwargs: Optional[Dict[str, Any]] = None,
    ) -> "Execution":
        """
        基于 moos 提供的上下文, 运行一段代码.
        :param code: 需要运行的代码.
        :param target: 指定上下文中一个变量名用来获取返回值. 如果为空, 则返回 None. 如果是一个方法, 则会运行它.
        :param local_args: 如果 args 不为 None, 则认为 target 是一个函数, 并从 locals 里指定参数
        :param local_kwargs: 类似 args
        :param args: 从外部注入的参数变量
        :param kwargs: 从外部注入的参数变量.
        :return: 根据 result_name 从 code 中获取返回值.
        :exception: any exception will be raised, handle them outside
        """
        from ghostos_moss.lifecycle import __moss_exec__
        self.update_executing_code(code)
        if self.__executing__:
            raise RuntimeError(f"Moss already executing")
        try:
            self.__executing__ = True
            compiled = self.module()
            fn = None
            with self.redirect_stdout():
                # 使用 module 自定义的 exec
                if hasattr(compiled, __moss_exec__.__name__):
                    fn = getattr(compiled, __moss_exec__.__name__)
                if fn is None:
                    fn = __moss_exec__
                # 使用系统默认的 exec
                return fn(
                    self,
                    target=target,
                    code=code,
                    local_args=local_args,
                    local_kwargs=local_kwargs,
                    args=args,
                    kwargs=kwargs,
                )
        finally:
            self.__executing__ = False

    @abstractmethod
    def update_executing_code(self, code: Optional[str] = None) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        """
        方便垃圾回收.
        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class Execution(NamedTuple):
    """
    result of the moss runtime execution.
    """
    returns: Any
    std_output: str
    pycontext: PyContext



from ghostos_container import Container
from ghostos_moss.abcd import (
    Moss, MossCompiler, MossRuntime, MossPrompter, Execution,
    AttrPrompts, Injection,
    MOSS_VALUE_NAME, MOSS_TYPE_NAME, MOSS_HIDDEN_MARK, MOSS_HIDDEN_UNMARK,
)
from ghostos_moss.moss_impl import DefaultMOSSProvider
from ghostos_moss.testsuite import MossTestSuite
from ghostos_moss.pycontext import PyContext
from ghostos_moss.exports import Exporter
from ghostos_moss.magics import __is_subclass__, __is_instance__, MagicPrompter
from ghostos_moss.prompts import PromptAbleClass, PromptAbleObj

__all__ = [
    # abstract contracts
    Moss, MossCompiler, MossRuntime, MossPrompter, Execution,
    # constants
    MOSS_VALUE_NAME, MOSS_TYPE_NAME, MOSS_HIDDEN_MARK, MOSS_HIDDEN_UNMARK,
    # types
    AttrPrompts,
    # pycontext related
    PyContext,
    # testing
    DefaultMOSSProvider,
    MossTestSuite,

    Exporter,  # useful to exports values in group, and other module will reflect them in moss_imported_attrs_prompt
    'moss_container',
    'moss_test_suite',

    # magic prompters
    __is_subclass__,
    __is_instance__,
    MagicPrompter,

    PromptAbleClass, PromptAbleObj,
]


def moss_container() -> Container:
    """
    test container for Moss
    """
    from ghostos.contracts.modules import DefaultModulesProvider
    container = Container()
    container.register(DefaultMOSSProvider())
    container.register(DefaultModulesProvider())
    return container


def moss_test_suite() -> MossTestSuite:
    """
    return a MossTestSuite
    """
    container = moss_container()
    return MossTestSuite(container)



from typing import TYPE_CHECKING
from typing import Optional, List, Dict, Any

if TYPE_CHECKING:
    from ghostos_moss.prompts import AttrPrompts
    from ghostos_moss.abcd import MossPrompter, Execution, MossRuntime, MossCompiler, Moss

"""
这个文件提供了 MOSS 生命周期的关键方法, 每一个都是可选的.
这样一个 Module 要在 MOSS 中运行时, 如果定义了这些方法或类, 就会替换掉默认生命周期里的相关方法. 
"""

__all__ = [
    '__moss_compile__',  # prepare moss compiler, handle dependencies register
    '__moss_compiled__',  # when moss instance is compiled
    '__moss_attr_prompts__',  # generate custom local attr prompt
    '__moss_module_prompt__',  # define module prompt
    '__moss_exec__',  # execute the generated code attach to the module
]

# todo: 整体重命名一遍魔术方法. 魔术方法并不是一个比较好的实现, 太依赖用户的知识了.


def __moss_compile__(compiler: "MossCompiler") -> "MossCompiler":
    """
    从 compile 中获取 MOSSRuntime, 并对它进行初始化.
    可选的魔术方法. 如果定义的话, MOSS 执行 compile 的阶段会默认执行它.

    主要解决各种注入方面的需求:
    compiler.inject()
    compiler.register()
    """
    return compiler


def __moss_compiled__(moss: "Moss") -> None:
    """
    当 Moss 对象被编译完成时, 回调的 hook
    :param moss:
    :return:
    """
    return


def __moss_attr_prompts__() -> "AttrPrompts":
    """
    生成本地变量生成的 prompt.
    可选的魔术方法.

    系统本身会反射代码的本地变量, 为它们中的一部分添加 prompt.
    默认的反射方法见 ghostos.moss.prompts.prompts.py 文件.

    而这个方法则可以替代或追加必要的 prompt, 优先于系统生成的反射.
    还有一些在 <moss-hide></moss-hide> 标记内定义的代码, 想要在 prompt 里呈现, 也可以在这个方法里定义.

    推荐使用 ghostos.moss.prompts 模块下提供的各种反射方法.
    :returns: Iterable[Tuple[name, prompt]] . 其中 name 只是为了去重.
    """
    return []


def __moss_module_prompt__(prompter: "MossPrompter", attr_prompts: Optional[Dict[str, str]] = None) -> str:
    """
    使用 MOSS Runtime 生成 prompt 的方法.
    可选的魔术方法. 定义的话, runtime.moss_context_prompt 实际上会使用这个方法.

    这个方法生成的 Prompt, 会用来描述当前文件, 其中包含了注入的 MOSS 类和 moss 实例.
    """
    from ghostos_moss.utils import escape_string_quotes

    # 获取原始的代码.
    origin_code = prompter.get_source_code(exclude_hide_code=True)

    # 基于 origin code 生成关于这些变量的 prompt.
    attrs_prompt_str = prompter.dump_imported_prompt(attr_prompts=attr_prompts)
    code_prompt_part = ""
    if attrs_prompt_str:
        # 这部分变量的描述, 放到一个 string 里表示不污染当前上下文.
        attrs_prompt_str = escape_string_quotes(attrs_prompt_str, '"""')
        code_prompt_part = f'''

# more details about some module attrs above, are list below (quoted by <attr></attr>):
"""
{attrs_prompt_str}
"""
'''

    # 生成完整的 prompt. 预计 MOSS 的描述已经在上下文里了.
    prompt = f"""
{origin_code}
{code_prompt_part}
"""

    return prompt


def __moss_exec__(
        runtime: "MossRuntime",
        *,
        target: str,
        code: "Optional[str]" = None,
        local_args: "Optional[List[str]]" = None,
        local_kwargs: "Optional[Dict[str, Any]]" = None,
        args: Optional[List[Any]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
) -> "Execution":
    """
    基于 MOSS Runtime 执行一段代码, 并且调用目标方法或返回目标值.
    :param runtime: moss runtime
    :param code: 会通过 execute
    :param target: 指定一个上下文中的变量, 如果是 callable 则执行它, 如果不是 callable 则直接返回.
    :param local_args: 为 target 准备的 *args 参数, 这里只需要指定上下文中的变量名.
    :param local_kwargs: 为 target 准备的 **kwargs 参数, 这里需要指定 argument_name => module_attr_name
    :param args: 从外部注入的参数变量
    :param kwargs: 从外部注入的参数变量.
    """
    from typing import Callable
    from ghostos_moss.abcd import Execution
    pycontext = runtime.dump_pycontext()
    pycontext.execute_code = code
    pycontext.executed = False

    local_values = runtime.locals()
    # 注意使用 runtime.exec_ctx 包裹有副作用的调用.
    if code:
        filename = pycontext.module if pycontext.module is not None else "<MOSS>"
        compiled = compile(code, filename=filename, mode='exec')
        exec(compiled, local_values)

    if target not in local_values:
        raise NotImplementedError(f"target `{target}` not implemented")

    target_module_attr = local_values.get(target, None)

    # 如果定义了参数, 就必须是 callable 方法.
    has_args = local_args is not None or local_kwargs is not None or args is not None or kwargs is not None
    if isinstance(target_module_attr, Callable):
        real_args = []
        real_kwargs = {}
        if local_args:
            for attr_name in local_args:
                if attr_name not in local_values:
                    raise AttributeError(f"module has no attribute `{attr_name}` for `{target}`")
                arg_val = local_values.get(attr_name, None)
                real_args.append(arg_val)
        if args:
            real_args.extend(args)
        if local_kwargs:
            for argument_name, attr_name in local_kwargs.items():
                if attr_name not in local_values:
                    raise AttributeError(f"module has no attribute {attr_name} for `{target}`")
                real_kwargs[argument_name] = local_values.get(attr_name, None)
        if kwargs:
            real_kwargs.update(kwargs)
        # 注意使用 runtime.exec_ctx 包裹有副作用的调用.
        with runtime.redirect_stdout():
            returns = target_module_attr(*real_args, **real_kwargs)
    elif has_args:
        raise TypeError(f"target '{target}' value '{target_module_attr}' is not callable")
    else:
        returns = target_module_attr
    std_output = runtime.dump_std_output()
    pycontext = runtime.dump_pycontext()
    pycontext.executed = True
    return Execution(returns, std_output, pycontext)
